/*
   ----------------------------------------------------------------------------
   PasswordMaskEF.cc
   Author: Ken Mayer
   Date..: June 30, 2015
   Description:
   Based on Password.cc in the dUFLP, this control ONLY handles the 
   mask character, it does *not* deal with processing the password,
   checking for valid/invalid passwords or other.
   
   In order to display the wanted character, this will use the
   Windows font: WingDings, and is set to display a solid dot,
   rather than the asterisk, which is common for a lot of password
   mask type controls. You could switch it out to display other
   symbols, such as the diamond shape, using the same font:
       chr( 108 ) Solid Dot
       chr( 116 ) diamond shape
       chr( 117 ) slightly fatter diamond shape
       chr( 118 ) diamond made of smaller diamonds ("star bullet" in Word)
       chr( 173 ) eight pointed star
   And of course you could try a wide variety of other symbols ...
   
   The password that is actually typed by the user is stored
   in the custom property: enteredPassword, and when
   checking it or using it, this is what you need to use.
   
   In the isValidChar method of the control, you can define
   additional characters that might be valid in a password.
   Many websites and services now allow for symbols such as #, $, etc.
   See the method if what you need is not already there, you
   can add it easily enough.
   
   Usage:
      set procedure to :dUFLP:PasswordMaskEF.cc
      // in the form designer, place an instance
      // of the PasswordMaskEF entryfield onto the form
      // DO NOT DATALINK IT! The value will not
      // be correct ... If you wish to work with
      // a DataLinked password, you should use
      // :dUFLP:Password.cc!
      // When checking for the password entered by
      // the user:
      cPass = form.PasswordMaskEF1.enteredPassword
      
   NOTE:
   Jan Hoelterling pointed out to me after I had posted this control
   in the newsgroups, that there is a simple way of handling all this --
   you can search teh web for "Password Font" and find a few fonts
   that simply replace the symbol displayed with a specific other symbol.
   The only concern I would have is installing a font on the users' 
   machine, but that can be done through Windows calls, or Inno Setup.
   The choice is yours. :)
   ----------------------------------------------------------------------------
*/

#define BACKSPACE_KEY     8     // -- ASCII value of Backspace key
#define DELETE_KEY        127   // -- ASCII value of Delete key
#define ENTER_KEY         13	  // -- ASCII value of Enter key
class PasswordMaskEF( oParent ) of Entryfield( oParent ) custom

   with( this )
      metric      := 6 // Pixels
      value       := ""
      maxLength   := 30
      height      := 22
      width       := 105
      fontName    := "WingDings"
      onOpen      := CLASS::ONOPEN
      key         := CLASS::KEY
   endwith
   // custom properties:
   this.maskChar           = chr(108) // solid dot
   // this is the value you must look for, 
   // rather than the "value" property ...
   this.enteredPassword = ""
   
   function OnOpen
      // Declare the CheckSelect API function prototype
      if Type("CheckSelect") # "FP"
         extern CLONG CheckSelect(CHANDLE, CUINT, ;
            CPTR CLONG, CPTR CLONG) user32 from "SendMessageA"
      endif

      // Custom properties
      if type("this.enteredPassword") = "U"
         this.enteredPassword = ""
      endif
   return
   
   function Key(nChar, nPosition)
      // Handles keys entered in the password entryfield
      private enteredChar, returnValue

      // Get the character positions of the selected text (if any)
      local SelectStartPos, SelectEndPos
      SelectStartPos = 0
      SelectEndPos = 0
      CheckSelect(this.hwnd,0xB0,SelectStartPos,SelectEndPos)

      enteredChar = chr(nChar)
      returnValue = true                // By default output whatever key was typed
      do case                           // Check for keys that modify the value
         case nChar = BACKSPACE_KEY
            // if the text is not selected, delete the previous character,
            // if the text is selected, delete the characters corresponding
            // to the selection range
            if SelectStartPos == 0 AND SelectEndPos == 0
               this.enteredPassword = ;
                  stuff(this.enteredPassword, nPosition - 1, 1, "")
            else
               this.enteredPassword = ;
                  stuff(this.enteredPassword, SelectStartPos + 1, ;
                  SelectEndPos - SelectStartPos, "")
            endif
         case nChar = DELETE_KEY
            // if the text is not selected, delete the current character,
            // if the text is selected, delete the characters corresponding
            // to the selection range
            if SelectStartPos == 0 AND SelectEndPos == 0
               this.enteredPassword = ;
                  stuff(this.enteredPassword, nPosition, 1, "")
            else
               this.enteredPassword = ;
                  stuff(this.enteredPassword, SelectStartPos + 1, ;
                  SelectEndPos - SelectStartPos, "")
            endif
         otherwise
            if CLASS::IsValidChar(enteredChar)    // Check if alphanumeric
               // if the text is not selected, insert the character entered,
               // if the text is selected, overwrite the characters
               // corresponding to the selection range with the character
               // entered
               if SelectStartPos == 0 AND SelectEndPos == 0
                  this.enteredPassword = ;
                     stuff(this.enteredPassword, nPosition, 1, enteredChar)
               else
                  this.enteredPassword = ;
                     stuff(this.enteredPassword, SelectStartPos + 1, ;
                     SelectEndPos - SelectStartPos, enteredChar)
               endif
               returnValue = asc( this.maskChar ) // Output camouflage character
            endif
      endcase
   return returnValue

   Procedure IsValidChar(char)
      // Make sure entered key is alphanumeric
      private returnValue

      do case
         case isalpha(char)                        // Letter?
            returnValue = true
         case char >= "0" and char <= "9"          // Digit?
            returnValue = true
         // if you need to allow others, add them here:
         case char $ "!@#$%^&*()_-+"               // Special characters
            returnValue = true
         otherwise                                 // Invalid?
            returnValue = false
      endcase
   return returnValue
endclass