/*
   --------------------------------------------------------
   Filename.....: appFrameSize.cc
   ClassName....: appFrameSize
   Version......: 5.xxxx.
   Purpose......: Non-Visual object to manage the framewin
                  position with a minimum height and
                  minimum width.
   Programmer...: Rick Miller http://www.treturn.com/dbase
   Date.........: September 02, 2004.
   Notes........: See Notes below.
   Written for..: dBASE PLUS, dB2K, and VdB 7.x.
   Rev. History.: See Revisions below.
   Dependencies.: :dUFLP:applicationIni.cc
   Calls........: see x_InitExterns method for API calls.
   Called by....: Any.
   Usage........: set procedure to appFrameSize.cc additive
   Example......: see How to use below.
   Sample.......: see appFrameSize.wfm.
   --------------------------------------------------------
   Notes:
   --------------------------------------------------------
      Always call the Release() method with:
         _app.frameSize.release()
         when finished with this object.
      This object automatically attaches
         itself to _app as _app.frameSize.
      Protected methods begin with "x_".
      Protected properties begin with "~".
   --------------------------------------------------------
   How to use:
   --------------------------------------------------------
      When starting an application.
         Local oRef
         oRef  =  new appFrameSize()

         // when changing the userIni property ?
         oRef.readIni()

         // use or change any properties.

      Methods used during the application.
         _app.frameSize.Center()
         _app.frameSize.FillDesktop()
         _app.frameSize.Maximize()
         _app.frameSize.Refresh()
         _app.frameSize.SetPosition()

      When ending the application.
         _app.frameSize.save()
         _app.frameSize.release()

   --------------------------------------------------------
   Properties:    Default:             Used for:
   --------------------------------------------------------
      height          ?                as read from userIni.
                                       height.
      left            ?                as read from userIni.
                                       left position.
      minHeight      300               minimum height.
      minWidth       400               minimum width.
      top             ?                as read from userIni.
                                       top position.
      userIni        _app.iniFile folder +
                     or
                     _dbwinhome folder +
                     userName + .ini   file to store settings.
      width           ?                as read from userIni.
      windowState     ?                as read from userIni.
                                       0/2 (Normal/Maximized).
   --------------------------------------------------------
   Methods:
   --------------------------------------------------------
      Center()
         1) call to insure valid property values.
         2) assign _app.frameWin.windowState =  0
         3) move _app.frameWin to the center of the
            screen using this.left, this.top,
            this.width, this.height.
      FillDesktop()
         1) call to insure valid property values.
         2) assign _app.frameWin.windowState =  0
         3) position _app.frameWin to the full
            size of the desktop client area.
      getAppIni()
         1) returns the application ini filename.
      Maximize()
         1) maximize the framewin.
         2) set the windowState property to 2.
      ReadIni()
         1) reads from userIni and assigns property values.
            [Desktop]
            FrameLeft
            FrameHeight
            FrameTop
            FrameWidth
            FrameMaximized
      Refresh()
         1) assign property values using the
            current frameWin windowState and position.
         return true/false (success/fail).
      Release()
         1) release object.
         2) close this procedure file.
      Save()
         read the current frameWin position and write the
            position to the userIni file in 1 of 3 ways.
         1) when _app.frameWin.windowState = 0
            The current _app.frameWin position is
               placed into property values that
               will be a visible position.
            Write to userIni.
            [Desktop]
            FrameLeft=this.Left
            FrameHeight=this.Height
            FrameTop=this.Top
            FrameWidth=this.Width
            FrameMaximized=0
            Maximized=0
            Position=this.left this.top
                     this.Left + this.width
                     this.top + this.Height
         2) when _app.frameWin.windowState = 1
            No writing to userIni.
         3) when _app.frameWin.windowState = 2
            Write to userIni.
            [Desktop]
            FrameMaximized=2
         return true/false (success/fail).
      SetPosition([bool])
         1) when windowState = 0, or passing a true.
            1) call to insure valid property values.
            2) assign _app.frameWin.windowState =  0
            3) move _app.frameWin to this.left this.top
               this.width this.height position.
         2) when windowState = 2
            1) maximize frameWin.
   --------------------------------------------------------
   Revisions:
   --------------------------------------------------------
      February 18, 2011
         1) default ini file adapted to following
            Windows security model and a standard User
            with the use of applicationIni.cc.
         2) added support for 2 monitors.
      March 24, 2005
         1) modified for use with non-BDE apps.
      March 21, 2005
         1) updated for dUFLP release.
         2) added getAppIni method for PLUS users.
      March 17, 2005
         1) modified x_CalcProperties method so that
            any property values saved will be
            within the viewable area of the desktop.
         2) added a parameter for SetPosition method.
         3) renamed ShowProps method to Inspect.
         4) added Maximize method.
      March 15, 2005
         1) added a file flush when calling save().
         2) added a return value to the save() method.
      February 10, 2005
         1) renamed all external functions.
      January 01, 2005
         1) verified external functions.
*/
#define  APPFRAMESIZE_DESKTOP       "Desktop"
#define  APPFRAMESIZE_PUBLISHERINFO "Rick Miller"
#define  APPFRAMESIZE_VERSIONINFO   "11.0218"
#define  MAX_USERNAME               256   // max user name length.
#define  SPI_GETWORKAREA            48    // working screen area.
#define  SM_CXSCREEN                0
#define  SM_CYSCREEN                1
#define  SM_CXVIRTUALSCREEN         78
#define  SM_CYVIRTUALSCREEN         79
#define  SM_CMONITORS               80
#define  SM_CXFRAME                 32
//---------------------------------------------------------------//
//          constructor.
//---------------------------------------------------------------//
class appFrameSize
   this.className =  "appFrameSize"
   this.minHeight =  300   // minimum frameWin height.
   this.minWidth  =  400   // minimum frameWin width.

   protect  ~appIni, x_CalcProperties, x_CloseProcedure,;
            x_FlushFileBuffers, x_GetDesktopArea, x_GetFrame,;
            x_GetLimits, x_GetRectValue, x_GetUserIni,;
            x_InitExterns, x_ReadIniInt, x_Save, x_SetState,;
            x_ValidateProperties, x_WriteIniInt, x_WriteIniString

   this.publisher    =  APPFRAMESIZE_PUBLISHERINFO
   this.versioninfo  =  APPFRAMESIZE_VERSIONINFO
   class::x_InitExterns()
   try
      set procedure to applicationIni.cc additive
   catch(exception e)
      try
         set procedure to :dUFLP:applicationIni.cc additive
      catch(exception e)
      endtry
   endtry
   this.~appIni   =  new applicationIni()
   this.userIni   =  class::x_GetUserIni()
   class::ReadIni()
   _app.frameSize =  this
   //------------------------------------------------------------//
   //       end of constructor   -  methods below.
   //------------------------------------------------------------//
   // center _app.frameWin with
   // the current property values in normal state.
   function Center
      class::x_ForceFrame()
      // call to insure a move can take place.
      class::x_ValidateProperties(true)
      // re-position the frameWin.
      RMM_MoveWindow(int(_app.frameWin.hwnd),;
               int((this.~Width - this.width) / 2 + this.~Left),;
               int((this.~Height - this.height) / 2 + this.~Top),;
               int(this.width), int(this.height), true)
   return

   //------------------------------------------------------------//
   // expand _app.frameWin to fill the desktop in normal state.
   function FillDesktop
      class::x_ForceFrame()
      // call to insure a move can take place.
      class::x_ValidateProperties(true)
      // re-position the frameWin.
      RMM_MoveWindow(int(_app.frameWin.hwnd),;
               int(this.~Left), int(this.~Top),;
               int(this.~Width), int(this.~Height), true)
   return

   //------------------------------------------------------------//
   // return the application.ini filename.
   function getAppIni
   return   this.~appIni.getFilename(true)

   //------------------------------------------------------------//
   // display properties in the inspector.
   function Inspect
      inspect(_app.frameSize)
   return

   //------------------------------------------------------------//
   // maximize the frameWin.
   function Maximize
      class::x_SetState(2)
   return   iif(this.windowState == 2, true, false)

   //------------------------------------------------------------//
   // assign property values from the userIni file.
   function ReadIni
      // read values from ini file.
      this.left   =  class::x_ReadIniInt("FrameLeft")
      this.top    =  class::x_ReadIniInt("FrameTop")
      this.height =  class::x_ReadIniInt("FrameHeight")
      this.width  =  class::x_ReadIniInt("FrameWidth")
      this.windowState  =  class::x_ReadIniInt("FrameMaximized")
   return

   //------------------------------------------------------------//
   // assign property values from the
   // current frameWin position and windowState.
   function Refresh
      Local bRet
      bRet  =  true
      class::x_ForceFrame()
      if _app.frameWin.windowState == 0
         Local cBuff, nHwnd
         nHwnd =  int(_app.frameWin.hwnd)
         cBuff =  space(8)
         if RMM_GetWindowRect(nHwnd, cBuff)
            Local nBott, nRite
            this.left   =  class::x_GetRectValue(cBuff, 0)
            this.top    =  class::x_GetRectValue(cBuff, 4)
            nRite =  class::x_GetRectValue(cBuff, 8)
            nBott =  class::x_GetRectValue(cBuff, 12)
            this.width  =  int(nRite - this.left)
            this.height =  int(nBott - this.top)
         endif
         this.windowState  := 0
      elseif _app.frameWin.windowState == 2
         this.windowState  := 2
      else
         bRet  := false
      endif
   return   iif(bRet, true, false)

   //------------------------------------------------------------//
   // null the _app.frameSize reference.
   // destroy this object.
   // call to remove this file from the procedure stack.
   function Release
      if type("this.~appIni") == "O"
         this.~appIni.close()
         this.~appIni   := null
      endif
      try
         _app.frameSize  =  null
         release object this
         class::x_CloseProcedure(program(1))
      catch(exception e)
      endtry
   return

   //------------------------------------------------------------//
   // refresh the property values.
   // validate the property values.
   // write property values to userIni.
   function Save
      Local bRet, nStat
      bRet  =  false
      nStat =  class::x_GetFrame()
      if nStat == 0
         // windowState == 0
         class::x_GetLimits()
         class::x_CalcProperties()
         class::x_WriteIniInt("FrameLeft",    this.left)
         class::x_WriteIniInt("FrameHeight",  this.height)
         class::x_WriteIniInt("FrameTop",     this.top)
         class::x_WriteIniInt("FrameWidth",   this.width)
         class::x_WriteIniInt("FrameMaximized", 0)
         class::x_WriteIniString("Position",;
                  ""  + int(this.left) +;
                  " " + int(this.top) +;
                  " " + int(this.left + this.width) +;
                  " " + int(this.top  + this.height))
         if class::x_WriteIniInt("Maximized", 0)
            bRet  := class::x_FlushFileBuffers()
         endif
      elseif nStat == 2
         if class::x_WriteIniInt("FrameMaximized", 2)
            bRet  := class::x_FlushFileBuffers()
         endif
      endif
   return   iif(bRet, true, false)

   //------------------------------------------------------------//
   // set the postion of the frameWin using the properties or
   // force a windowState of normal and
   // position the framewin with property values.
   function SetPosition(bForce)
      class::x_ForceFrame()
      if this.windowState == 0 or bForce
         // call to insure a move can take place.
         class::x_ValidateProperties()
         // re-position the frameWin.
         RMM_MoveWindow( ;
                  _app.frameWin.hwnd, this.left, this.top,;
                  this.width, this.height, true)
      else
         // call to maximize the frameWin.
         class::x_SetState(2)
      endif
   return

   //------------------------------------------------------------//
   // when necessary, adjust property values
   // that will insure the entire frameWin
   // will be viewable.
   function x_CalcProperties

      if this.width < this.minWidth
         // not enough width.
         this.width  := int(this.minWidth)
      endif
      if this.width > this.~Width
         // too much width.
         this.width  := int(this.~Width)
      endif

      if this.height < this.minHeight
         // not enough height.
         this.height := int(this.minHeight)
      endif
      if this.height > this.~Height
         // too much height.
         this.height := int(this.~Height)
      endif

      if this.left + this.width > this.~Width
         // too far right.
         // display width on screen.
         this.left   := int(this.~Width - this.width)
      endif
      if this.left < this.~Left
         // too far left.
         this.left   := int(this.~Left)
      endif

      if this.top + this.height > this.~Height
         // too low.
         // display minimum height on screen.
         this.top := int(this.~Height - this.height)
      endif
      if this.top < this.~Top
         // too high.
         this.top := int(this.~Top)
      endif

   return

   //------------------------------------------------------------//
   function x_CheckMonitor(bCenter)
         Local    nPvx, nPvy, nPx, nPy
         // get primary screen size.
         // primary screen width.
         nPx   =  RMM_GetSystemMetrics(SM_CXSCREEN)
         // primary screen height.
         nPy   =  RMM_GetSystemMetrics(SM_CYSCREEN)
         // get the total virtual size.
         // the virtual width is the total of the 2 screens.
         // virtual width.
         nPvx  =  RMM_GetSystemMetrics(SM_CXVIRTUALSCREEN)
         // the virtual height is the larger of the 2 screens.
         nPvy  =  RMM_GetSystemMetrics(SM_CYVIRTUALSCREEN)
         // determine which screen is in use.
         // get hWndParent position.
         if this.left > nPx - RMM_GetSystemMetrics(SM_CXFRAME)
            // using the 2nd screen.
            // the 2nd screen is positioned to the right.
            this.~Left     := nPx
            if bCenter
               this.~Width    := int(nPvx - nPx)
            else
               this.~Width    := int(nPvx)
            endif
            this.~Height   := int(nPvy)
         endif
   return

   //------------------------------------------------------------//
   // remove a file from the procedure stack.
   function x_CloseProcedure(cProc)
      try
         close procedure ("" + cProc)
      catch(exception e)
      endtry
   return

   //------------------------------------------------------------//
   // when necessary open an mdi form out of view
   // to force the frameWin to display.
   function x_ForceFrame
      if not type("_app.frameWin") == "O" or ;
                  _app.frameWin.hwnd == 0
         Local oForm
         oForm =  new form()
         oForm.top   =  100
         oForm.open()
         // destroy the mdi form.
         oForm.release()
         oForm := null
      endif
   return

   //------------------------------------------------------------//
   // write any userIni buffered data to disk.
   function x_FlushFileBuffers
      Local bRet, cFile, oFile
      bRet  =  false
      if this.userIni.indexof("\") == -1
         // windows default ini folder.
         cFile =  GetEnv("windir") + "\" + this.userIni
      else
         cFile =  "" + this.userIni
      endif
      oFile =  new file()
      try
         oFile.open(cFile, "W")
         if oFile.handle == -1
            bRet  := false
         else
            bRet  := RMM_FlushFileBuffers(int(oFile.handle))
         endif
         oFile.close()
      catch(exception e)
         // a read-only file will fail
         // with an exception when using open("W").
      endtry
      oFile := null
   return   iif(bRet, true, false)

   //------------------------------------------------------------//
   // return an object with the desktop client area.
   function x_GetDesktopArea
      Local cBuff, o
      o  =  new object()
      cBuff =  space(8)
      RMM_GetClientDesktopRect(SPI_GETWORKAREA, 0, cBuff, 0)
      o.left   =  class::x_GetRectValue(cBuff, 0)
      o.top    =  class::x_GetRectValue(cBuff, 4)
      o.right  =  class::x_GetRectValue(cBuff, 8)
      o.bottom =  class::x_GetRectValue(cBuff, 12)
   return   o

   //------------------------------------------------------------//
   // call to assign properties from the framewin position
   // when the framewin.windowState == 0 (normal).
   function x_GetFrame
      Local nRet
      nRet  =  int(_app.framewin.windowState)
      if nRet == 0
         // the frameWin is in a normal state.
         if not class::Refresh()
            // the frameWin is minimized.
            nRet  := 1
         endif
      endif
   return   nRet

   //------------------------------------------------------------//
   // assign the desktop sceen size limits to protected properties.
   function x_GetLimits(bCenter)
      oTemp =  class::x_GetDesktopArea()
      if type("this.~Left") == "U"
         protect  ~Height, ~Left, ~Top, ~Width
      endif
      this.~Left     =  int(oTemp.left)
      this.~Top      =  int(oTemp.top)
      this.~Height   =  int(oTemp.bottom - oTemp.top)
      this.~Width    =  int(oTemp.right - oTemp.left)
      release object oTemp
      oTemp := null
      if RMM_GetSystemMetrics(SM_CMONITORS) > 1
         this.x_CheckMonitor(bCenter)
      endif
   return

   //------------------------------------------------------------//
   // read a portion of a Rect structure.
   function x_GetRectValue( cBuff, nByte )
      Local nRet
      nRet  =  cBuff.getbyte(nByte) + ;
               BitLshift(cBuff.getbyte(nByte + 1), 8) + ;
               BitLshift(cBuff.getbyte(nByte + 2), 16) + ;
               BitLshift(cBuff.getbyte(nByte + 3), 24)
      if BitSet(nRet, 31)
         nRet  := - BitNot(nRet)
      endif
   return   int(nRet)

   //------------------------------------------------------------//
   // assign a default userIni property.
   function x_GetUserIni
      Local cFold, cUser, sBuff, sLen
      cFold =  this.~appIni.getPath()
      sBuff =  replicate(chr(0), MAX_USERNAME + 1)
      cUser =  "User"
      sLen  =  replicate(chr(0), 4)
      sLen.setByte(0, bitand(MAX_USERNAME, 0xFF))
      sLen.setByte(1, bitand(bitzrshift(MAX_USERNAME,  8), 0xFF))
      sLen.setByte(2, bitand(bitzrshift(MAX_USERNAME, 16), 0xFF))
      sLen.setByte(3, bitand(bitzrshift(MAX_USERNAME, 24), 0xFF))
      if RMM_GetUserName(sBuff, sLen)
         cUser := sBuff.left(sBuff.indexOf(chr(0)))
      endif
   return   (cFold + "\" + cUser + ".ini")

   //------------------------------------------------------------//
   // initialize external functions.
   function x_InitExterns
      if not type("RMM_FlushFileBuffers") == "FP"
         extern   CLOGICAL ;
                  RMM_FlushFileBuffers(;
                  CHANDLE) ;
                  kernel32 from "FlushFileBuffers"
      endif
      if not type("RMM_GetClientDesktopRect") == "FP"
         extern   CLOGICAL ;
                  RMM_GetClientDesktopRect(;
                  CUINT,;
                  CUINT,;
                  CPTR,;
                  CUINT) ;
                  user32 from "SystemParametersInfoA"
      endif
      if not type("RMM_GetSystemMetrics") == "FP"
         extern   CINT ;
                  RMM_GetSystemMetrics(;
                  CINT) ;
                  user32 from "GetSystemMetrics"
      endif
      if not type("RMM_GetUserName") == "FP"
         extern   CLOGICAL ;
                  RMM_GetUserName(;
                  CSTRING,;
                  CPTR) ;
                  advapi32.dll from "GetUserNameA"
      endif
      if not type("RMM_GetWindowRect") == "FP"
         extern   CLOGICAL ;
                  RMM_GetWindowRect(;
                  CHANDLE,;
                  CPTR) ;
                  user32 from "GetWindowRect"
      endif
      if not type("RMM_MoveWindow") == "FP"
         extern   CLOGICAL ;
                  RMM_MoveWindow(;
                  CHANDLE,;
                  CINT,;
                  CINT,;
                  CINT,;
                  CINT,;
                  CLOGICAL) ;
                  user32 from "MoveWindow"
      endif
   return

   //------------------------------------------------------------//
   // read an integer from an ini file.
   function x_ReadIniInt(cEntry)
   return   int(val(;
            this.~appIni.getValue(;
            APPFRAMESIZE_DESKTOP,;
            cEntry,;
            this.userIni)))

   //------------------------------------------------------------//
   // toggle the framewin windowState if necessary.
   function x_SetState(nStat)
      if not _app.framewin.windowState == nStat
         _app.framewin.windowState  := nStat
         this.windowState  := int(nStat)
      endif
   return

   //------------------------------------------------------------//
   function x_ValidateProperties(bCenter)
      // refresh the screen size limits.
      class::x_GetLimits(bCenter)
      // call to valididate property position values.
      class::x_CalcProperties()
      // call to make sure windowState is normal.
      class::x_SetState(0)
   return

   //------------------------------------------------------------//
   // write a value to the ini file.
   function x_WriteIniInt(cEntry, nVal)
   return   this.~appIni.setValue(;
            APPFRAMESIZE_DESKTOP,;
            cEntry,;
            "" + int(0 + nVal),;
            this.userIni)

   //------------------------------------------------------------//
   // write a value to the ini file.
   function x_WriteIniString(cEntry, cVal)
   return   this.~appIni.setValue(;
            APPFRAMESIZE_DESKTOP,;
            cEntry,;
            "" + cVal,;
            this.userIni)

endclass
//---------------------------------------------------------------//
//          end of appFrameSize.
//---------------------------------------------------------------//

