CLASS setArray 
/* 
    ------------------------------------------------------------------
    Programmer..: Roland Wingerter
    Date........: 11 February 2003
    Written for.: dBase Plus

    Notes.......: Some methods for working with arrays which are sets.
                  A set is an unordered collection of distinct elements
                  (i.e. an array which has no duplicates).  

                  All methods in this class can be applied to arrays as
                  well. Most methods however are designed specifically
                  to work with sets. They could be kept in a separate class,
                  while the others (like append, array2set) could be 
                  added to the more general arrayex class. 

                  Note that unlike the methods in arrayex, currently all 
                  methods in this class must be called with parameters. 
                  (I personally prefer it that way, but this could be
                  changed). Means some more work, though ;-)

                  Restrictions: 
                  - This class does not handle multi-dimensional arrays. 
                  - It does not handle adequately arrays/sets that contain 
                    other arrays/sets. 

    ------------------------------------------------------------------

    Instantiation:
 
      set procedure to :dUFLP:rwArray.cc additive
      oSetArray = new setArray()

    Methods:
 
       addNoDup
         Add Element to Array if no duplicate, i.e.
         element is not already contained in array.
 
       append  // array method, not specifically for sets 
         Append two arrays, by adding all elements from Array2 to the 
         end of Array1.

       appendNoDup  // array method, not specifically for sets 
         Append Array2 without duplicates. In other words,
         append only elements from Array2 that are not already 
         members of Array1. 

       array2Set  // array method, not specifically for sets
         Transform array into set by removing duplicates.
         (A set is an array that contains unique elements.)

       calcSimilarity
         Calculate similarity between two sets in percent.
           
       elimDup  // array method, not specifically for sets 
         Eliminate duplicates from array.

       getComplement
         The complement of set1 contains all members of set2
         which are not members of set1.

       getDifference
         Difference is defined as the set of elements 
         contained in set1, but not in set2.

       getIntersection
         Intersection is defined as the set of unique elements that are 
         contained in both sets compared. 

       getRelation 
         Classify how set1 relates to set2
         (Subset, Disjunct, Identical,..)

       getSymmDiff
         Symmetrical difference is defined as the set 
         containing all elements that belong to only one of the 
         two sets given (exclusive or).

       getUnion
         The Union of two sets comprises all (unique) elements from set1 
         and set2. 

       hasDuplicates
         Test if array has duplicates. 

       isProperSubset
         Compares two sets and tests if Set1 is a proper subset of Set2.
         A proper Subset is a Subset that is smaller in size
         than its superset. 

       isProperSuperset
         Counterpart (reciprocal relation) to isProperSubset().  

       isSet  // array method, not specifically for sets 
         Test if a given array is a set. A set is defined as 
         a collection of unique objects, i.e. duplicate values 
         are not allowed. An empty array is also a set. 

       isSubset
         Compares two sets and tests if Set1 is a subset of Set2. 
         A subset is a defined as a set whose elements are 
         contained in the superset. 

       isValidType
         Checks argument type (auxiliary method)

       passed arrays
         Checks argument type (auxiliary method) 
   ----------------------------------------------------------------
*/

   FUNCTION array2set( oArray )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Transform array into set by removing duplicates.
                   A set is an array that contains unique elements. 

                   Note that the elements in the set will be ordered. 
                   (The order of elements in a set is not important). 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: elimDup(), isValidType()
     Usage.......: oSet = new setArray()
                   oSet.array2set( Array )
     Example.....: a1 = {"butter","milk","butter","honey"}
                   r = oSet.array2set( a1 ) 
                   ? r.size // 3 {"butter","honey","milk"}
     Returns.....: An array containing unique elements (i.e. a set)
     Parameter...: Array that will be transformed. 
     ------------------------------------------------------------------
   */
      local aResultSet

      if not class::isValidType( oArray, "A", "array2Set" )
         return false
      else
         if oArray.Size = 0
         	aResultSet = oArray   // empty array
         else
            // call method to remove duplicates 
            aResultSet = class::elimDup( oArray )
         endif
      endif

   return aResultSet
   // -- EoM: array2Set()


   FUNCTION getUnion( oArray1, oArray2 )	
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Union comprises all unique elements from Array1 
                   and Array2. In other words, union consists of all elements 
                   that are contained in either Array1 or Array2. 
                   The result is a set, i.e. it does not 
                   contain duplicates. 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: appendNoDup, elimDup()
     Usage.......: oSet = new setArray()
                   oSet.getUnion( Array, Array )
     Example.....: r = oSet.getUnion( {1,2},{'a','b'}) // r = {1,2,'a','b'}
     Returns.....: An array (set) containing all unique elements from 
                   Array1 (set1) _and_ Array2 (set2). 
     Parameters..: Array1, Array2
     ------------------------------------------------------------------
   */
   // First, elimDup() eliminates any duplicates from oArray1.
   // Then, appendnoDup() appends only new elements from oArray2
	return class::appendNoDup(class::elimDup(oArray1),oArray2)
   // -- EoM: getUnion()
                                                                                                   

   FUNCTION getIntersection( oArray1, oArray2 )	
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Intersection: set of unique elements that are 
                   contained in both arrays.

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: 
     Usage.......: oSet = new setArray()
                   oSet.getIntersection( Array1, Array2 )
     Example.....: r = oSet.getIntersection( {1,2,3},{3,4,5}) // r = {3}
                   a1 = {"boys","like,"girls"} 
                   a2 = {"girls","like","pets"} 
                   r = oSet.getIntersection( a1, a2 ) 
                   ? r.size // 2 
                   ? r[1] = "like"
                   ? r[2] = "girls"
     Returns.....: An array containing all elements that are
                   contained in both arrays. 
     Parameters..: Array1, Array2
     ------------------------------------------------------------------ 
   */
      local aIntersection, lValid
      private lsetExact

     // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "getIntersection" )
      if not lValid ; return -1 ; endif   // error 
                                                                                                   
      lExact = set("Exact")
      set exact on

      // create empty array
      aIntersection = new array(0)

      // For each member of Array1 
      for i=1 to oArray1.Size
         // If member of Array2
         if oArray2.scan( oArray1[i] ) # 0
            // And not member of result array
            // (only add unique values)
            if aIntersection.scan( oArray1[i] ) = 0
               // Common element, add to result
               aIntersection.Add( oArray1[i] )
            endif
         endif
      next i

      set exact &lExact

   return aIntersection
   // -- EoM: getIntersection


   FUNCTION getDifference( oArray1, oArray2 ) 
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Difference is defined as the set of elements 
                   contained in Array1, but not in Array2.
                   You may think of this as Array1 "minus" Array2. 
                   The result does not contain duplicates.

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: passedArrays()
     Usage.......: oSet = new setArray()
                   oSet.getDifference( Array, Array )
     Example.....: r = oSet.getDifference( {1,2,3,4},{3,4}) // r = {1,2}
     Returns.....: An array containing the elements from 
                   oArray1 not contained in oArray2. 
     Parameters..: Array1, Array2
     ------------------------------------------------------------------
   */
      local aDifference, lValid
      private lExact

     // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "getDifference" )
      if not lValid ; return -1 ; endif   // error 

      lExact = set("Exact")
      set exact on

      // Create empty result array                                                                                                   
      aDifference = new array(0)

      // For each member of Array1
      for i=1 to oArray1.Size
         // If not member of Array2
         if oArray2.scan( oArray1[i] ) = 0 
            // And not member of result set
            if aDifference.scan( oArray1[i] ) = 0 
               // Add new element to result
               aDifference.Add( oArray1[i] )
            endif
         endif
      next i
                                                                                                   
      set exact &lexact
                                                                                  
   return aDifference
   // -- EoM: getDifference

 
   FUNCTION getSymmDiff( oArray1, oArray2 )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Symmetrical difference is defined as the set 
                   containing all elements that belong to only one of the 
                   two arrays/sets (exclusive or). 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: getDifference(), getUnion()
     Usage.......: oSet = new setArray()
                   oSet.getSymmDiff( Array, Array )
     Example.....: r = oSet.getSymmDiff( {1,2,'a','b'},{1,2,'X','Y'}) 
                   // r = {'a','b','X','Y'} 
     Returns.....: An array containing the elements contained 
                   in Array1 or in Array2. 
     Parameters..: Array1, Array2
     ------------------------------------------------------------------
   */
      local aSymmDiff, aDiff1, aDiff2, lValid

     // Make sure arguments passed are arrays 
		lValid = class::passedArrays( oArray1, oArray2, "getSymmDiff" )
      if not lValid ; return -1 ; endif   // error 
                                                                    
      aDiff1 	= class::getDifference( oArray1, oArray2 ) 
      aDiff2 	= class::getDifference( oArray2, oArray1) 

      // We use getUnion() instead of Append() in order to remove duplicates
      aSymmDiff = class::getUnion( aDiff1, aDiff2 )
                                                                     
   return aSymmDiff
   // -- EoM: getSymDiff


   FUNCTION isIdentical( oArray1, oArray2 )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Two sets are identical if they have the same size
                   and contain the same elements, not regarding their
                   order. 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: getIntersection(), passedArrays()
     Usage.......: oSet = new setArray()
                   oSet.isIdentical( Array, Array )
     Example.....: ? oSet.isIdentical({1,2,3},{3,2,1}) // true
     Returns.....: An array containing the elements contained 
                   in Array1 or in Array2. 
     Parameters..: Array1, Array2
     ------------------------------------------------------------------
   */
      local lIdentical, lValid

     // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "isIdentical" )
      if not lValid ; return -1 ; endif   // error 
                                                                    
      if oArray1.Size # oArray2.Size
         // identical array must have same size
         lIdentical = false
      else
         // Both arrays have the same number of elements
         // Get common elements contained in both arrays 
         aIntersection = class::getIntersection( oArray1, oArray2 )
         lIdentical = iif( oArray1.Size = aIntersection.Size, true, false  )
      endif                                                                     
   return lIdentical
   // -- EoM: isIdentical


 	FUNCTION hasDuplicates( oArray )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Test array for duplicates

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: isValidType()
     Usage.......: oSet = new setArray()
                   oSet.hasDuplicates( Array )
     Example.....: ? oSet.hasDuplicates({1,1,2,3,3}) // true
     Returns.....: true or false
     Parameters..: Array to be tested
     ------------------------------------------------------------------
   */
 	   local lRetVal, aTemp

      // Make sure argument passed is an array
      if not class::isValidType(oArray, "A", "hasDuplicates" ) 
         // if it is not an array, we can't tell if it has duplicates
         return -1   // error code
      endif

      // Check to see if array has duplicates
      if oArray.Size <= 1   // 0 or 1 elements
         lRetVal = false    // no duplicates possible
      else
         // Array size >= 2
         lRetVal = false   // provide default value

         // Create sorted copy of array
         aTemp   		= new array( oArray.Size )
         acopy(oArray,aTemp)
         aTemp.Sort()

         // For each element in temporary Array                                                                                       
         for i=2 to aTemp.Size
            // Check if element equals previous element
 	         if aTemp[i] == aTemp[i-1]
               // found duplicate value!
 	            lRetVal=true
 	            exit
            endif
         next i
      endif
                                                                                                   
   return lRetVal
   // -- EoM: hasDuplictes
                                                                                                   

   FUNCTION isSet( oArray )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Test if array is a set.
                   A set is a collection of unique objects, i.e.
                   duplicate values are not allowed. 
                   Note that an empty array is also a set. 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: hasDuplicates(), isValidType()
     Usage.......: oSet = new setArray()
                   oSet.isSet( Array )
     Example.....: ? oSet.isSet({1,2,3}) // true
                   ? oSet.isSet({3,3,3}) // false
     Returns.....: true or false
     Parameters..: Array to be tested
     ------------------------------------------------------------------
   */
      local lSet

      // Make sure argument passed is an array
      if not class::isValidType( oArray, "A", "isSet" )
         // if it's not an array, it can't be a set! 
         return false
      else
         if oArray.Size = 0   
            lSet = true   // empty arrays are sets!
         else 
            // It is a set if it does not contain duplicates
            lSet = not class::hasDuplicates(oArray)
         endif
      endif
   
   return lSet
   // -- EoM: isSet
 

   FUNCTION isSubset( oArray1, oArray2 )	// Subset, oArray
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Test if Array1 is a subset of Array2.
                   Returns true if all elements of Array1 (subset) are 
                   contained in Array2.
                   In sets, the sequence of elements in both arrays 
                   does not matter, e.g. {1,2,3} is a subset of {4,3,2,1}. 

                   Note that a subset can be the same size as its superset. 
                   See also isProperSubset(), where the subset must be smaller.

     Written for.: dBase Plus                         
     Rev. History: None
     Calls.......: getDifference(), passedArrays()
     Usage.......: oSet = new setArray()
                   oSet.isSubset( Array, Array )
     Example.....: ? oSet.isSubset({1,2},{1,2,3,4}) // true
                   ? oSet.isSubset({1,2},{3,4,5}) // false
                   ? oSet.isSubset({1,2},{1,2}) // true !
     Returns.....: true or false
     Parameters..: Array to be tested
     ------------------------------------------------------------------
   */
      local aDifference, lSubset, lValid

      // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "isSubset" )
      if not lValid ; return -1 ; endif   // error 

      if oArray1.Size > oArray2.Size
         // a subset cannot be bigger in size
         lSubset = false
      else
         // Get elements contained in Array1 but not in Array2
         aDifference = class::getDifference( oArray1, oArray2 )
         // If Array1 has different elements from Array2, it is not a subset 
         lSubset 		= iif( aDifference.Size=0, true, false )
      endif
                                                   
   return lSubset
   // -- EoM: isSubset
 

   FUNCTION isProperSubset( oArray1, oArray2 )   // Subset, Array 
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Test if Array1 is a proper subset of Array2.
                   Returns true if Array1 (subset) is smaller and 
                   contains all elements of Array2. 
                   Same as isSubset(), but Array1 must be smaller.

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: isSubset(), passedArrays()
     Usage.......: oSet = new setArray()
                   oSet.isProperSubset( Array, Array )
     Example.....: ? oSet.isProperSubset({1,2},{1,2,3,4}) // true
                   ? oSet.isProperSubset({1,2},{4,3,2,1}) // true
                   ? oSet.isProperSubset({1,2},{1,2}) // false
     Returns.....: true or false
     Parameters..: Array to be tested
     ------------------------------------------------------------------
   */
      local lRetVal, lValid

     // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "isProperSubset" )
      if not lValid ; return -1 ; endif   // error 

      if oArray1.size < oArray2.size ;
         and class::isSubset(oArray1,oArray2)
         lRetVal = true
 		else
 	      lRetVal = false
      endif
                                                                                                   
   return lRetVal
   // -- EoM: isProperSubset


   FUNCTION isProperSuperset ( oArray1, oArray2 ) // Superset, Array
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Test if Array1 is a proper superset of Array2.
                   This is the reciprocal relation to isProperSubset()

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: isProperSubset()
     Usage.......: oSet = new setArray()
                   oSet.isProperSuperset( Array, Array )
     Example.....: ? oSet.isProperSuperset({1,2,3,4},{1,2}) // true
                   ? oSet.isProperSuperset({1,2},{1,2}) // false
     Returns.....: true or false
     Parameters..: Array to be tested
     ------------------------------------------------------------------
   */
   return class::isProperSubset ( oArray2, oArray1 )
   // -- EoM: isProperSuperset [sic]


   FUNCTION getRelation( oArray1, oArray2 )   
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Type of relation expresses how Array1 relates to Array2. 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: getIntersection(), passedArrays()
     Usage.......: oSet = new setArray()
                   oSet.getRelation( Array, Array )

     Examples....: r=oSet.getrelation({1,2,3},{'X','Y'}) // r="DISJUNCT"
                   r=oSet.getrelation({1,2,3},{1,3,2})   // r="IDENTITY"
                   r=oSet.getrelation({1,2},{3,2,1})     // r="SUBSET"
                   r=oSet.getrelation({3,2,1},{1,2})     // r="SUPERSET"
                   r=oSet.getrelation({1,2,3},{3,4,5})   // r="INTERSECT"

     Returns.....: A string characterizing the relation:
                   "DISJUNCT"  - no common elements
                   "IDENTITY"  - only common elements
                   "SUBSET"    - Array1 is (proper!) subset of Array2
                   "SUPERSET"  - Array1 is (proper!) superset of Array2
                   "INTERSECT" - Arrays have common elements and 
                                 each has at least one of its own
     Parameters..: Arrays to be compared
     ------------------------------------------------------------------
   */
      local aIntersection, cRelType, lValid

      // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "getRelation" )
      if not lValid ; return -1 ; endif   // error 

      aIntersection = class::getIntersection( oArray1, oArray2 )

      // aIntersection is empty
      if aIntersection.Size = 0
         do case
            // Both input arrays are empty 
            case oArray1.Size = 0 and oArray2.Size = 0 
               cRelType = "IDENTITY"
            // Array1 is empty
            case oArray1.Size = 0
                  cRelType = "SUBSET"
            // Array2 is empty
            case oArray2.Size = 0
               cRelType = "SUPERSET"

            otherwise
               cRelType = "DISJUNCT"
         endcase

      else // aIntersection is not empty
         do case
            // Intersect size equals both sizes
            case aIntersection.Size = oArray1.Size ;
                  and aIntersection.Size = oArray2.Size
               cRelType = "IDENTITY"

            // Intersect size equals one size
            case aIntersection.Size = oArray1.Size ; 
                  or aIntersection.Size = oArray2.Size
               if oArray1.Size > oArray2.Size
                  cRelType = "SUPERSET"
               else
                 cRelType = "SUBSET"
               endif

            otherwise
               cRelType = "INTERSECT"
         endcase
      endif
                                                                                                   
   return cRelType
   // -- EoM: getRelation


   FUNCTION calcSimilarity( oArray1, oArray2 )   
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Calculates similarity between two sets in percent.

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: getRelation(); getIntersection(), getUnion()
                   passedArrays()

     Usage.......: oSet = new setArray()
                   oSet.calcSimilarity( Array, Array )

     Example.....: ? oSet.calcSimilarity({1,2},{2,1})     // 100
                   ? oSet.calcSimilarity({1,2},{1})       // 50
                   ? oSet.calcSimilarity({1,2},{2,3})     // 33
                   aHobbies = {'computers','sports','books'}
                   aOccupation = {'programming','computers','software'}
                   ? oSet.calcSimilarity(aHobbies,aOccupation) //  20
                   ? oSet.calcSimilarity(aHobbies,aOccupation) > 0 // true

     Returns.....: Integer between 0 and 100, representing 
                   similarity of the two arrays in percent
     Parameters..: Array to be tested
     ------------------------------------------------------------------
   */
      local aIntersection, aUnion, nSimilar, lValid

     // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "calcSimilarity" )
      if not lValid ; return -1 ; endif   // error 

      cRelType = class::getRelation( oArray1, oArray2 )

      do case
         // Maximum similarity: Identical sets are 100% similar
         case left(upper(cRelType),5) = "IDENT" 
            nSimilar = 100

         // No similarity: Disjunct sets are not similar at all
         case left(upper(cRelType),5) = "DISJU" 
            nSimilar = 0

         otherwise
            // Get intersection and union to calculate the similarity 
            aIntersection = class::getIntersection( oArray1, oArray2 )
            aUnion        = class::getUnion( oArray1, oArray2 )

            // How many percent of all elements (=union) are the same? 
            nSimilar      = int(round(aIntersection.Size * 100 ;
                                      / aUnion.Size,0 )) 
      endcase
                                                                                     
   return nSimilar
   // -- EoM: calcSimilarity

    
   FUNCTION getComplement( oArray1, oArray2 )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Definition of Complement: 
                   If Array1 is subset of Array2, 
     					 aComplement contains all members of Array2
                   which are not members of Array1.
                   In other words, "If Array1 is a subset of Array2, 
                   what else is there in Array2'? 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: isSubset(), getDifference(), passedArrays()
     Usage.......: oSet = new setArray()
                   oSet.getComplement( Array, Array )
     Example.....: r = oSet.getComplement({1,2},{2,1,'X'}) // r = {'X'}
                   r = oSet.getComplement({1,2},{2,1,'X','Y'}) // r = {'X','Y'}
     Returns.....: Array with Complement
     Parameters..: oArray1, oArray2
     ------------------------------------------------------------------
   */                                                                       
      local aComplement, lValid

     // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, getComplement )
      if not lValid ; return -1 ; endif   // error 

      if class::isSubset( oArray1, oArray2 ) = true
         aComplement = class::getdifference( oArray2, oArray1 )
      else
         aComplement = new array(0)   // empty array
      endif
                                                                                              
   return aComplement
   // -- EoM: getComplement

                                             
   FUNCTION append( oArray1, oArray2 )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Append two arrays, by adding all elements from Array2 to the 
                   end of Array1. 

                   See also: setArray::appendNoDup, which appends only 
                   unique values from Array2. 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: passedArrays()
     Usage.......: oSet = new setArray()
                   oSet.append( Array, Array )
     Example.....: r = oSet.append({1,2},{2,1,'X}) // r = {1,2,2,1,'X'}
     Returns.....: Array 
     Parameters..: Array1 - first array
                   Array2 - array you want to append
     ------------------------------------------------------------------
   */                                                                       
      local aResult, lValid

      // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "Append" )
      if not lValid ; return -1 ; endif   // error 

      // Copy Array1 to aResult
      aResult = new array(oArray1.Size)
      if oArray1.Size > 0
         aCopy(oArray1,aResult)
      endif

      // Add all elements from Array2
      for i=1 to oArray2.Size
         aResult.Add( oArray2[i] )
      next i
   return aResult
   // -- EoM: append                                                                                                   


   FUNCTION appendNoDup( oArray1, oArray2 )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Append Array2 without duplicates. In other words,
                   append only elements from Array2 that are not already 
                   members of Array1. 

                   Note: This leaves alone any duplicates in Array1. 
                   Use getUnion() if you want the result to contain 
                   only unique values. 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: passedArrays()
     Usage.......: oSet = new setArray()
                   oSet.appendNoDup( Array, Array )
     Example.....: r = oSet.appendNoDup({1,2},{2,1,'X}) // r = {1,2,'X'}
     Returns.....: Array 
     Parameters..: Array1 - first Array
                   Array2 - Array to append elements from. 
     ------------------------------------------------------------------
   */                                                                       
      local aResult, lValid
      private lExact

      // Make sure arguments passed are arrays
		lValid = class::passedArrays( oArray1, oArray2, "appendNoDup" )
      if not lValid ; return -1 ; endif   // error 

      // Need to set exact
      lExact = set("Exact")
      set exact on
                                                                                      
      // Copy Array1 to aResult
      aResult = new array(oArray1.Size)
      if oArray1.Size > 0
         acopy(oArray1,aResult)
      endif

      // For each element in Array2                                                                                   
      for i=1 to oArray2.Size
         // Add unique elements only 
         if aResult.Scan( oArray2[i] ) = 0
            aResult.Add( oArray2[i] )
         endif
      next i

      set exact &lExact 

   return aResult
   // -- EoM: appendNoDup


   FUNCTION addNoDup( oArray, xElement )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Add Element to Array if no duplicate, i.e.
                   element is not already contained in array.
                   Note that this does not return a new array, but
                   actually changes the original input array,  
                   (but leaves alone any duplicates in input array)

                   Warning: Do not use this to add arrays. You can add
                   arrays, but duplicate arrays will not be recognized. 

                   This method has some overhead when used in a loop,
                   because each time the parameter passed and the setting 
                   is checked. 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: isValidType()
     Usage.......: oSet = new setarray()
                   oSet.addNoDup(Array,Element)
     Example.....: r=oSet.addNoDup({1,2,3},3) // r={1,2,3}
                   r=oSet.addNoDup({1,2,3},4) // r={1,2,3,4}
                   a1={1,2,3} 
                   r=oSet.addNoDup(a1,4) // r={1,2,3,4}
                   ? a1.size // 4, since original input array is changed!
     Returns.....: Original array with element added. 
     Parameters..: oArray1 - Array; Element to be added
     ------------------------------------------------------------------
   */                                                                       
      local lValid
      private lExact

      // First parameter passed must be an array.
      // Second parameter could be anything or missing (which adds false)
      lValid = class::isValidType(oArray, "A", "addNoDup" )
      if not lValid ; return -1 ; endif   // error 

      lExact = set("Exact")
      set exact on
                                                                                                   
      // Add only elements not already contained in Array
      if oArray.Scan( xElement ) = 0
         oArray.Add( xElement )
      endif
                                                                                                   
      set exact &lExact
                                                                                  
   return oArray
   // -- EoM: addNoDup


   FUNCTION elimDup( oArray )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Eliminate duplicates from array.
                   Returns a set of unique elements in sorted order. 

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: isValidType()
     Usage.......: oSet = new setarray()
                   oSet.elimDup( Array )
     Example.....: r=oSet.elimDup({1,2,2) // r={1,2}
                   a1 = {'a','rose','is','a','rose'} 
                   ? oSet.elimDup(a1).Size // 3
                   r=oSet.elimDup(a1) // r={'a','is','rose'}
     Returns.....: Array containing unique values.
     Parameters..: Array from which you want to remove duplicates. 
     ------------------------------------------------------------------
   */                                                                       
      local aResultSet, aTemp

      // Check to make sure argument is an array
      if not class::isValidType(oArray, "A", "elimDup" )
         // If it's not an array, leave it alone. 
         // isValidType() has already displayed a message
      	return oArray
      endif

      if oArray.Size <= 1   // size 0 or 1
      	aResultSet = oArray   // no duplicates possible
      else
         
         // Create sorted copy of array
         // to allow faster discovery of duplicates
         aTemp   		= new array( oArray.Size )
         acopy(oArray,aTemp)
         aTemp.Sort()

         // Create empty result set
         aResultSet 	= new array(0)
         // Add first element 
         aResultSet.Add( aTemp[1] ) 

         // For all remaining elements
         for i=2 to aTemp.Size
            // Add element if not equal last element
            if aTemp[i] <> aResultSet[aResultSet.Size]
               aResultSet.Add( aTemp[i] )
            endif
         next i
      endif
                                                                                             
 	return aResultSet
   // -- EoM: elimDup


   FUNCTION isValidType
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Check if argument has correct type. 
                   If not, display error message.  
                   Auxiliary method.  

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: None
     Usage.......: --
     Example.....: See class methods
     Returns.....: true or false
     Parameters..: Argument, cRequiredType
     ------------------------------------------------------------------
   */                                                                       
      parameters Argument, cRequiredType, cMethodName

      local lValid                                       

      if type("Argument") # cRequiredType
         msgbox("Wrong argument type - Required type is " ;
            + cRequiredType, "Message from: " + cMethodname)
         lValid = false
      else
         lValid = true
      endif

   return lValid
   // -- EoM: isValidType


   FUNCTION passedArrays ( oArray1, oArray2, cMethodName )
   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: 11 February 2003

     Notes.......: Check if arguments provided are arrays. 
                   Auxiliary method.  

     Written for.: dBase Plus
     Rev. History: None
     Calls.......: None
     Usage.......: --
     Example.....: See class methods
     Returns.....: true or false
     Parameters..: Parameter1: First Array, 
                   Parameter2: Second Array
                   Parameter3: Name of calling Method
     ------------------------------------------------------------------
   */                                                                       
      local lValid1, lValid2

      // Check argument types
      lValid1 = class::isValidType(oArray1, "A", cMethodName )
      lValid2 = class::isValidType(oArray2, "A", cMethodName )

	return (lValid1 and lValid2)
   // -- EoM: passedArrays
      
ENDCLASS
// End of class: setArray                                                                                           

//------------------------------------------------------------------
// EOF: SETARRAY.CC
//------------------------------------------------------------------

